﻿using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;
using System.Collections;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Collections.ObjectModel;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;

using alib.Enumerable;
using alib.Debugging;
using alib.Wpf;

namespace xigt.app
{
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public sealed class ColumnDisplayInfo : DependencyObject
	{
		public static readonly DependencyProperty KeyProperty =
			DependencyProperty.RegisterAttached("Key", typeof(Guid), typeof(ColumnDisplayInfo), new PropertyMetadata(Guid.Empty));

		public static Guid GetKey(DependencyObject obj) { return (Guid)obj.GetValue(KeyProperty); }
		public static void SetKey(DependencyObject obj, Guid value) { obj.SetValue(KeyProperty, value); }
		public Guid Key
		{
			get { return (Guid)GetValue(KeyProperty); }
			set { SetValue(KeyProperty, value); }
		}

		public static readonly DependencyProperty WidthProperty =
			DependencyProperty.Register("Width", typeof(Double), typeof(ColumnDisplayInfo), new PropertyMetadata(Double.NaN,
				(o, e) =>
				{
					Debug.Print("{0}", (Double)e.NewValue);
				}, (d, o) =>
				{
					Double w;
					if (!Double.IsNaN(w = (Double)o) && w < 0.0)
						return 0.0;
					return o;
				}));

		public Double Width
		{
			get { return (Double)GetValue(WidthProperty); }
			set { SetValue(WidthProperty, value); }
		}

		public static readonly DependencyProperty IsVisibleProperty =
			DependencyProperty.Register("IsVisible", typeof(bool), typeof(ColumnDisplayInfo), new PropertyMetadata(true,
				(o, e) => ((ColumnDisplayInfo)o).Visibility = (bool)e.NewValue ? Visibility.Visible : Visibility.Collapsed));

		[DefaultValue(true)]
		public bool IsVisible
		{
			get { return (bool)GetValue(IsVisibleProperty); }
			set { SetValue(IsVisibleProperty, value); }
		}

		public static readonly DependencyProperty VisibilityProperty =
			DependencyProperty.Register("Visibility", typeof(Visibility), typeof(ColumnDisplayInfo), new PropertyMetadata(Visibility.Visible,
				(o, e) => ((ColumnDisplayInfo)o).IsVisible = (Visibility)e.NewValue == Visibility.Visible));

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public Visibility Visibility
		{
			get { return (Visibility)GetValue(VisibilityProperty); }
			set { SetValue(VisibilityProperty, value); }
		}

		public static readonly DependencyProperty IsSortKeyProperty =
			DependencyProperty.Register("IsSortKey", typeof(bool), typeof(ColumnDisplayInfo), new PropertyMetadata(false));

		[DefaultValue(false)]
		public bool IsSortKey
		{
			get { return (bool)GetValue(IsSortKeyProperty); }
			set { SetValue(IsSortKeyProperty, value); }
		}
	};

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public class GridViewColumnDisplayInfo : DependencyObject, IReadOnlyList<ColumnDisplayInfo>, IList<ColumnDisplayInfo>, IList
	{
		///////////////////////////////////////////////////////
		/// 
		public static readonly DependencyProperty DividerPositionProperty =
			DependencyProperty.Register("DividerPosition", typeof(Double), typeof(GridViewColumnDisplayInfo), new PropertyMetadata(230.0));

		public Double DividerPosition
		{
			get { return (Double)GetValue(DividerPositionProperty); }
			set { SetValue(DividerPositionProperty, value); }
		}

		///////////////////////////////////////////////////////
		/// 
		public static DependencyProperty ColumnWidthTotalProperty { get { return ColumnWidthTotalPropertyKey.DependencyProperty; } }

		static readonly DependencyPropertyKey ColumnWidthTotalPropertyKey =
			DependencyProperty.RegisterReadOnly("ColumnWidthTotal", typeof(Double), typeof(GridViewColumnDisplayInfo), new PropertyMetadata(0.0, null,
				(d, _) =>
				{
					var l = ((GridViewColumnDisplayInfo)d).list;
					Double w = 0.0;
					int i = l.Count;
					while (--i >= 0)
						w += l[i].Width;
					return w;
				}));

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public Double ColumnWidthTotal { get { return (Double)GetValue(ColumnWidthTotalProperty); } }

		///////////////////////////////////////////////////////
		/// 
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public int Count { get { return (int)GetValue(CountProperty); } }

		public static DependencyProperty CountProperty { get { return CountPropertyKey.DependencyProperty; } }

		static readonly DependencyPropertyKey CountPropertyKey =
			DependencyProperty.RegisterReadOnly("Count", typeof(int), typeof(GridViewColumnDisplayInfo), new PropertyMetadata(0, null,
				(d, _) => ((GridViewColumnDisplayInfo)d).list.Count));


		///////////////////////////////////////////////////////
		/// 
		public GridViewColumnDisplayInfo(IEnumerable<ColumnDisplayInfo> items)
		{
			this.list = new oc(this, items);
		}
		public GridViewColumnDisplayInfo()
			: this(null)
		{
		}

		sealed class oc : ObservableCollection<ColumnDisplayInfo>
		{
			static oc()
			{
				dpd_col_width = DependencyPropertyDescriptor.FromProperty(ColumnDisplayInfo.WidthProperty, typeof(ColumnDisplayInfo));
			}

			static DependencyPropertyDescriptor dpd_col_width;

			public oc(GridViewColumnDisplayInfo owner, IEnumerable<ColumnDisplayInfo> items)
				: base(items ?? alib.Collections.Collection<ColumnDisplayInfo>.None)
			{
				this.owner = owner;
			}
			readonly GridViewColumnDisplayInfo owner;

			protected override void RemoveItem(int ix)
			{
				var _old = base[ix];
				base.RemoveItem(ix);
				dpd_col_width.RemoveValueChanged(_old, column_width_change);
			}
			protected override void InsertItem(int ix, ColumnDisplayInfo _new)
			{
				dpd_col_width.AddValueChanged(_new, column_width_change);
				base.InsertItem(ix, _new);
			}
			protected override void SetItem(int ix, ColumnDisplayInfo _new)
			{
				var _old = base[ix];
				if (_old != _new)
				{
					dpd_col_width.AddValueChanged(_new, column_width_change);
					base.SetItem(ix, _new);
					dpd_col_width.RemoveValueChanged(_old, column_width_change);
				}
			}
			protected override void ClearItems()
			{
				base.ClearItems();
				foreach (var col in this)
					dpd_col_width.RemoveValueChanged(col, column_width_change);
			}

			protected override void OnPropertyChanged(PropertyChangedEventArgs e)
			{
				base.OnPropertyChanged(e);
				if (e.PropertyName == "Count")
					owner.CoerceValue(CountProperty);

				column_width_change(null, null);
			}
			
			void column_width_change(Object _, EventArgs __)
			{
				owner.CoerceValue(ColumnWidthTotalProperty);
			}
		};

		readonly oc list;

		public ColumnDisplayInfo this[int ix]
		{
			get { return list[ix]; }
			set { list[ix] = value; }
		}

		public void Add(ColumnDisplayInfo item) { list.Add(item); }

		public void Insert(int ix, ColumnDisplayInfo item) { list.Insert(ix, item); }

		public bool Remove(ColumnDisplayInfo item) { return list.Remove(item); }

		public void RemoveAt(int ix) { list.RemoveAt(ix); }

		public int IndexOf(ColumnDisplayInfo item) { return list.IndexOf(item); }

		public bool Contains(ColumnDisplayInfo item) { return list.Contains(item); }

		public void Clear() { list.Clear(); }

		public void CopyTo(ColumnDisplayInfo[] array, int arrayIndex) { list.CopyTo(array, arrayIndex); }

		public IEnumerator<ColumnDisplayInfo> GetEnumerator() { return list.GetEnumerator(); }

		IEnumerator IEnumerable.GetEnumerator() { return list.GetEnumerator(); }

		public bool IsReadOnly { get { return ((IList)list).IsReadOnly; } }

		Object IList.this[int ix]
		{
			get { return list[ix]; }
			set { list[ix] = (ColumnDisplayInfo)value; }
		}

		int IList.Add(object value)
		{
			list.Add((ColumnDisplayInfo)value);
			return Count - 1;
		}
		void IList.Insert(int index, object value) { list.Insert(index, (ColumnDisplayInfo)value); }
		void IList.Remove(object value) { list.Remove((ColumnDisplayInfo)value); }
		bool IList.Contains(object value) { return list.Contains((ColumnDisplayInfo)value); }
		int IList.IndexOf(object value) { return list.IndexOf((ColumnDisplayInfo)value); }
		void ICollection.CopyTo(Array array, int ix) { ((ICollection)list).CopyTo(array, ix); }
		bool IList.IsFixedSize { get { return ((IList)list).IsFixedSize; } }
		bool ICollection.IsSynchronized { get { return ((ICollection)list).IsSynchronized; } }
		Object ICollection.SyncRoot { get { return ((ICollection)list).SyncRoot; } }
	};

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	static class __util
	{
		public static RenderTargetBitmap GetImage(UIElement fe, Brush background = null, Size sz = default(Size), int dpi = 144)
		{
			if (sz.Width < alib.Math.math.ε || sz.Height < alib.Math.math.ε)
			{
				fe.Measure(util.infinite_size);
				sz = fe.DesiredSize; //VisualTreeHelper.GetContentBounds(fe).Size; //
			}

			DrawingVisual dv = new DrawingVisual();
			RenderOptions.SetEdgeMode(dv, EdgeMode.Aliased);

			using (DrawingContext ctx = dv.RenderOpen())
			{
				Rect r = new Rect(0, 0, sz.Width, sz.Height);
				if (background != null)
					ctx.DrawRectangle(background, null, r);

				VisualBrush br = new VisualBrush(fe);
				br.AutoLayoutContent = true;
				ctx.DrawRectangle(br, null, r);
			}

			Double f = dpi / 96.0;

			RenderTargetBitmap bitmap = new RenderTargetBitmap(
				(int)(sz.Width * f) + 1,
				(int)(sz.Height * f) + 1,
				dpi,
				dpi,
				PixelFormats.Pbgra32);
			bitmap.Render(dv);
			return bitmap;
		}
	};
}
