﻿using System;
using System.ComponentModel;
using System.Globalization;
using System.Collections.Generic;
using System.IO;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Markup;
using System.Windows.Media;
using System.Xml;
using System.Xaml;

using alib;
using alib.Array;
using alib.file;
using alib.Collections;
using alib.Debugging;
using alib.Wpf;

namespace xigt.app
{
	using Enum = System.Enum;

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public abstract class app_config_base : DependencyObject
	{
		protected static T _load<T>(String fn)
			where T : app_config_base, new()
		{
			T _new;

			if (App.FindConfigFile(ref fn))
			{
				_new = (T)App._load(fn);
			}
			else
			{
				_new = new T();
			}
			_new.filename = fn;
			return _new;
		}

		public void Save()
		{
			App.save_to_xaml(this, filename);
		}

		public String filename;
	};

	public enum StartupAction
	{
		None = 0,
		New = 1,
		ReloadLastFileOrNone = 2,
		ReloadLastFileOrNew = 3,
		ReloadLastSessionOrNone = 4,
		ReloadLastSessionOrNew = 5,
	};

	public enum LineEndingMode
	{
		Preserve = 1,
		Windows = 2,
		Unix = 3,
	};

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public class EqualsConverter<T> : MarkupExtension, IValueConverter
	{
		public EqualsConverter(T param)
		{
			this.selected_value = param;
			this.deselected_value = (T)Activator.CreateInstance(param.GetType());
		}
		readonly T selected_value;
		readonly T deselected_value;

		public override object ProvideValue(IServiceProvider sp) { return this; }

		public object Convert(object value, Type T, object parameter, CultureInfo culture)
		{
			return value.Equals(selected_value);
		}

		public object ConvertBack(object value, Type T, object parameter, CultureInfo culture)
		{
			return (bool)value ? selected_value : deselected_value;
		}
	};
	public sealed class EnumEquals : EqualsConverter<Enum>
	{
		public EnumEquals(Enum param)
			: base(param)
		{
		}
	};

	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public class Settings : app_config_base
	{
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		static Settings inst;
		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		public static Settings Instance
		{
			[DebuggerStepThrough]
			get
			{
				var _tmp = inst;
				if (_tmp == null)
					_tmp = inst = _load<Settings>("app-settings.xaml");
				return _tmp;
			}
		}

		public Settings()
		{
			this.col_info_corpora = new GridViewColumnDisplayInfo();
			this.col_info_corpus = new GridViewColumnDisplayInfo();

			this.xw_settings = new XmlWriterSettings
			{
				Indent = true,
				CloseOutput = true,
			};

			Reset();
		}

		public readonly XmlWriterSettings xw_settings;

		public void Reset()
		{
			col_info_corpora.Clear();

			col_info_corpus.Clear();

#if false
			col_info_corpora.Clear();
			foreach (var col in ((GridView)w_corpora.View).Columns)
				col_info_corpora.Add(new ColumnDisplayInfo
				{
					Key = ColumnDisplayInfo.GetKey(col),
					Width = col.ActualWidth,
				});

			col_info_corpus.Clear();
			foreach (var col in ((GridView)w_corpus.View).Columns)
				col_info_corpus.Add(new ColumnDisplayInfo
				{
					Key = ColumnDisplayInfo.GetKey(col),
					Width = col.ActualWidth,
				});
#endif

			this.tier_colors = new Dictionary<String, Color>
			{
#if false
				{ "odin", 0xFFFFD0.ToColor() },
				{ "translations", 0xFFD0E0.ToColor() },
				{ "phrases", 0xD0FFFF.ToColor() },
				{ "words", 0xFFFFE0.ToColor() },
				{ "morphemes", 0xFFE0FF.ToColor() },
				{ "glosses", 0xE0FFFF.ToColor() },
				{ "translation-words", 0xD0E0FF.ToColor() },
				{ "bilingual-alignments", 0xE0FFD0.ToColor() },
				{ "tokenization", 0xD0FFE0.ToColor() },
				{ "pos", 0xFFE0D0.ToColor() },
				{ "phrase-structure", 0xE0D0E0.ToColor() },
				{ "dependencies", 0xE0E0FF.ToColor() },
#endif
			};

			this.LastDirectory = Environment.CurrentDirectory;
			this.SessionFiles = new String[0];
			this.recent_files = new String[0];
			this.SelectedCorpusIndex = -1;

			ClearValue(EscapeKeyExitProperty);
			ClearValue(StartupActionProperty);
			ClearValue(SaveOnExitProperty);
			ClearValue(WindowStateProperty);
			ClearValue(ShowIdsProperty);
			ClearValue(GridSplitterKeyboardAccessProperty);

			CoerceValue(LineEndingModeProperty);

			ClearValue(EncodingNameProperty);
			CoerceValue(EncodingNameProperty);
		}

		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		/// 
		///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

		GridViewColumnDisplayInfo col_info_corpora;
		public GridViewColumnDisplayInfo CorporaColumnDisplayInfo
		{
			get { return col_info_corpora; }
			set { col_info_corpora = value; }
		}

		GridViewColumnDisplayInfo col_info_corpus;
		public GridViewColumnDisplayInfo CorpusColumnDisplayInfo
		{
			get { return col_info_corpus; }
			set { col_info_corpus = value; }
		}

		Dictionary<String, Color> tier_colors;
		public Dictionary<String, Color> TierColors
		{
			get { return tier_colors; }
			set { tier_colors = value; }
		}

		String[] recent_files;
		public String[] RecentFiles
		{
			get { return recent_files; }
			set { recent_files = value; }
		}
		public void RemoveRecentFile(String fn)
		{
			arr.RemoveAll(ref recent_files, fn);
		}
		public void SetRecentFile(String fn)
		{
			LastDirectory = Path.GetDirectoryName(fn);

			int ix;
			if ((ix = Array.IndexOf(recent_files, fn)) == 0)
				return;

			if (ix > 0)
				arr.RemoveAt(ref recent_files, ix);
			else if (recent_files.Length == 20)
				arr.RemoveLast(ref recent_files);

			arr.Prepend(ref recent_files, fn);
		}

		public String LastDirectory { get; set; }

		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		String[] _sess_files;
		public String[] SessionFiles
		{
			get { return _sess_files; }
			set { _sess_files = value; }
		}

		[DebuggerBrowsable(DebuggerBrowsableState.Never)]
		int _corpus_ix;
		public int SelectedCorpusIndex
		{
			get { return _corpus_ix; }
			set { _corpus_ix = value; }
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public String SelectedCorpusFilename
		{
			get { return (uint)_corpus_ix < (uint)_sess_files.Length ? _sess_files[_corpus_ix] : null; }
		}

		public static readonly DependencyProperty StartupActionProperty =
			DependencyProperty.Register("StartupAction", typeof(StartupAction), typeof(Settings), new PropertyMetadata(StartupAction.ReloadLastSessionOrNew));

		public StartupAction StartupAction
		{
			get { return (StartupAction)GetValue(StartupActionProperty); }
			set { SetValue(StartupActionProperty, value); }
		}

		public static readonly DependencyProperty LineEndingModeProperty =
			DependencyProperty.Register("LineEndingMode", typeof(LineEndingMode), typeof(Settings), new PropertyMetadata(LineEndingMode.Unix, null,
				(d, o) =>
				{
					var _xs = ((Settings)d).xw_settings;
					switch ((LineEndingMode)o)
					{
						case LineEndingMode.Preserve:
							_xs.NewLineHandling = NewLineHandling.None;
							break;
						case LineEndingMode.Unix:
							_xs.NewLineChars = "\n";
							_xs.NewLineHandling = NewLineHandling.Replace;
							break;
						case LineEndingMode.Windows:
							_xs.NewLineChars = "\r\n";
							_xs.NewLineHandling = NewLineHandling.Replace;
							break;
						default:
							return DependencyProperty.UnsetValue;
					}
					return o;
				}));

		public LineEndingMode LineEndingMode
		{
			get { return (LineEndingMode)GetValue(LineEndingModeProperty); }
			set { SetValue(LineEndingModeProperty, value); }
		}

		public static readonly DependencyProperty EncodingNameProperty =
			DependencyProperty.Register("EncodingName", typeof(String), typeof(Settings), new PropertyMetadata("utf-8", null,
				(d, o) =>
				{
					EncodingInfo _tmp;
					var _this = (Settings)d;
					var _new = (String)o;

					if ((_tmp = _this.EncodingInfo) != null && _new.Equals(_tmp.Name, StringComparison.OrdinalIgnoreCase))
						return _tmp.Name;

					var rg = Encoding.GetEncodings();
					for (int i = rg.Length; --i >= 0; )
						if (_new.Equals((_tmp = rg[i]).Name, StringComparison.OrdinalIgnoreCase))
						{
							_this.EncodingInfo = _tmp;
							return _tmp.Name;
						}
					return DependencyProperty.UnsetValue;
				}));

		public String EncodingName
		{
			get { return (String)GetValue(EncodingNameProperty); }
			set { SetValue(EncodingNameProperty, value); }
		}

		public static readonly DependencyProperty EncodingInfoProperty =
			DependencyProperty.Register("EncodingInfo", typeof(EncodingInfo), typeof(Settings), new PropertyMetadata(null,
				(o, e) =>
				{
					var _this = (Settings)o;
					_this.SetValue(EncodingNameProperty, ((EncodingInfo)e.NewValue).Name);
					_this.CoerceValue(CanBOMPropertyKey.DependencyProperty);
					_this.CoerceValue(EncodingPropertyKey.DependencyProperty);
				}));

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public EncodingInfo EncodingInfo
		{
			get { return (EncodingInfo)GetValue(EncodingInfoProperty); }
			set { SetValue(EncodingInfoProperty, value); }
		}

		static readonly DependencyPropertyKey CanBOMPropertyKey =
			DependencyProperty.RegisterReadOnly("CanBOM", typeof(bool), typeof(Settings), new PropertyMetadata(false,
				(o, e) => ((Settings)o).CoerceValue(BOMProperty),
				(d, _) =>
				{
					var s_enc = ((Settings)d).EncodingInfo.Name;
					return s_enc.StartsWith("utf-") && s_enc != "utf-7";
				}));

		public bool CanBOM
		{
			get { return (bool)GetValue(CanBOMPropertyKey.DependencyProperty); }
		}

		public static readonly DependencyProperty BOMProperty =
			DependencyProperty.Register("BOM", typeof(bool), typeof(Settings), new PropertyMetadata(false,
				(o, e) => ((Settings)o).CoerceValue(EncodingPropertyKey.DependencyProperty),
				(d, o) => (bool)o && ((Settings)d).CanBOM));

		public bool BOM
		{
			get { return (bool)GetValue(BOMProperty); }
			set { SetValue(BOMProperty, value); }
		}

		static readonly DependencyPropertyKey EncodingPropertyKey =
			DependencyProperty.RegisterReadOnly("Encoding", typeof(Encoding), typeof(Settings), new PropertyMetadata(null,
			(o, e) => ((Settings)o).xw_settings.Encoding = (Encoding)e.NewValue,
			(d, _) =>
			{
				var _this = (Settings)d;
				var _inf = _this.EncodingInfo;
				Encoding _tmp;
				switch (_inf.Name)
				{
					case "utf-32":
						_tmp = new UTF32Encoding(false, _this.BOM);
						break;
					case "utf-16":
						_tmp = new UnicodeEncoding(false, _this.BOM);
						break;
					case "utf-8":
						_tmp = new UTF8Encoding(_this.BOM);
						break;
					default:
						_tmp = _inf.GetEncoding();
						break;
				}

				if ((_tmp.GetPreamble().Length > 0) != (_this.BOM))
					throw not.valid;
				return _tmp;
			}));

		public Encoding Encoding
		{
			get { return (Encoding)GetValue(EncodingPropertyKey.DependencyProperty); }
		}

		public static readonly DependencyProperty GridSplitterKeyboardAccessProperty =
			DependencyProperty.Register("GridSplitterKeyboardAccess", typeof(bool), typeof(Settings), new PropertyMetadata(false));

		public bool GridSplitterKeyboardAccess
		{
			get { return (bool)GetValue(GridSplitterKeyboardAccessProperty); }
			set { SetValue(GridSplitterKeyboardAccessProperty, value); }
		}

		public static readonly DependencyProperty EscapeKeyExitProperty =
			DependencyProperty.Register("EscapeKeyExit", typeof(bool), typeof(Settings), new PropertyMetadata(false));

		public bool EscapeKeyExit
		{
			get { return (bool)GetValue(EscapeKeyExitProperty); }
			set { SetValue(EscapeKeyExitProperty, value); }
		}

		public static readonly DependencyProperty SaveOnExitProperty =
			DependencyProperty.Register("SaveOnExit", typeof(bool), typeof(Settings), new PropertyMetadata(true));

		public bool SaveOnExit
		{
			get { return (bool)GetValue(SaveOnExitProperty); }
			set { SetValue(SaveOnExitProperty, value); }
		}

		public static readonly DependencyProperty WindowStateProperty =
			DependencyProperty.Register("WindowState", typeof(WindowState), typeof(Settings), new PropertyMetadata(WindowState.Normal));

		public WindowState WindowState
		{
			get { return (WindowState)GetValue(WindowStateProperty); }
			set { SetValue(WindowStateProperty, value); }
		}

		public static readonly DependencyProperty ShowIdsProperty =
			DependencyProperty.Register("ShowIds", typeof(bool?), typeof(Settings), new PropertyMetadata(true));

		public bool? ShowIds
		{
			get { return (bool?)GetValue(ShowIdsProperty); }
			set { SetValue(ShowIdsProperty, value); }
		}

		public static readonly DependencyProperty ShowAlignRefsProperty =
			DependencyProperty.Register("ShowAlignRefs", typeof(bool?), typeof(Settings), new PropertyMetadata(true));

		public bool? ShowAlignRefs
		{
			get { return (bool?)GetValue(ShowAlignRefsProperty); }
			set { SetValue(ShowAlignRefsProperty, value); }
		}

		public static readonly DependencyProperty ShowContentRefsProperty =
			DependencyProperty.Register("ShowContentRefs", typeof(bool?), typeof(Settings), new PropertyMetadata(true));

		public bool? ShowContentRefs
		{
			get { return (bool?)GetValue(ShowContentRefsProperty); }
			set { SetValue(ShowContentRefsProperty, value); }
		}

		public static readonly DependencyProperty ShowLineNumbersProperty =
			DependencyProperty.Register("ShowLineNumbers", typeof(bool?), typeof(Settings), new PropertyMetadata(null));

		public bool? ShowLineNumbers
		{
			get { return (bool?)GetValue(ShowLineNumbersProperty); }
			set { SetValue(ShowLineNumbersProperty, value); }
		}

		public static readonly DependencyProperty ShowTagsProperty =
			DependencyProperty.Register("ShowTags", typeof(bool?), typeof(Settings), new PropertyMetadata(null));

		public bool? ShowTags
		{
			get { return (bool?)GetValue(ShowTagsProperty); }
			set { SetValue(ShowTagsProperty, value); }
		}
	};

	[MarkupExtensionReturnType(typeof(xigt.app.Settings))]
	public sealed class SettingsExtension : MarkupExtension
	{
		public SettingsExtension() { }
		public SettingsExtension(PropertyPath path) { this.path = path; }

		PropertyPath path;
		public PropertyPath Path
		{
			get { return path; }
			set { path = value; }
		}
		public override Object ProvideValue(IServiceProvider sp)
		{
			IProvideValueTarget pvt;
			if (path != null && (pvt = (IProvideValueTarget)sp.GetService(typeof(IProvideValueTarget))) != null)
				((Binding)pvt.TargetObject).Path = path;

			return xigt.app.Settings.Instance;
		}
	};
}
