﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Windows;
using System.Text;
using System.Windows.Markup;
using System.Threading;
using System.Threading.Tasks;
using System.Xaml;
using System.Xml;

using alib;
using alib.Debugging;
using alib.String;
using alib.Enumerable;

namespace xigt
{
	///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	/// 
	public class xigt_corpus : _host<corpora, igt>
	{
		static String coerce_file(String fn)
		{
			if (fn != null)
			{
				fn = Path.GetFullPath(fn);
				if (!System.IO.Directory.Exists(Path.GetDirectoryName(fn)))
					fn = null;
				else if (File.Exists(fn))
					fn = new Uri(fn).LocalPath;
			}
			return fn;
		}

		public static xigt_corpus Load(String fn)
		{
			fn = coerce_file(fn);

			if (!File.Exists(fn))
				throw not.valid;

			var xd = new corpus_xmldoc();

			using (var xr = XmlReader.Create(fn, new XmlReaderSettings
			{
				IgnoreComments = false,
			}))
			{
				xd.Load(xr);
			}
			var xc = xd.DocumentElement;

			xc.FullPath = fn;

			xc.OnXmlLoaded();

			xc.Modified = false;
			return xc;
		}

		public static xigt_corpus New()
		{
			var xd = new corpus_xmldoc();
			var xc = (xigt_corpus)xd.CreateElement("xigt-corpus");
			xd.AppendChild(xc);
			return xc;
		}

		public bool Save(XmlWriterSettings settings)
		{
			var fn = FullPath;
			if (fn == null)
				throw not.valid;

			if (!NeedSave)
			{
				Debug.Print("DID NOT SAVE [{0}] because is did not seem to be modified", fn);
				return true;
			}

			using (var xw = XmlWriter.Create(fn, settings))
			{
				var xd = base.OwnerDocument;
				var _sig = xd.CreateComment(String.Format(" written by XigtEdit {0} ", DateTime.Now));
				xd.InsertBefore(_sig, xd.DocumentElement);
				xd.Save(xw);
				xd.RemoveChild(_sig);
			}
			Debug.Print("saved [{0}]", fn);

			//CoerceValue(ExistsPropertyKey.DependencyProperty);
			Modified = false;
			return true;
		}

		public igt InsertIgt(int ix)
		{
			var e_igt = (igt)OwnerDocument.CreateElement("igt");
			this.Insert(ix, e_igt);
			e_igt.id = make_igt_id();
			return e_igt;
		}

		String make_igt_id()
		{
			var alpha = new HashSet<String>();
			ulong i, max_id = 0;
			foreach (var igt in this)
			{
				var s = igt.id;
				if (ulong.TryParse(s.Where(ch => '0' <= ch && ch <= '9').NewString(), out i) && i > max_id)
					max_id = i;
				var sa = s.Where(ch => ch < '0' || ch > '9').NewString();
				if (sa.Length > 0)
					alpha.Add(sa);
			}
			String s_id;
			var pfx = alpha.FirstOrDefault() ?? "i";
			do
			{
				++max_id;
				s_id = pfx + max_id;
			}
			while (this.ContainsId(s_id));
			return s_id;
		}

		public xigt_corpus(String pfx, String name, String s_ns, corpus_xmldoc xd)
			: base(pfx, name, s_ns, xd)
		{
			xd.NodeInserted += xd_NodeChanged;
			xd.NodeRemoved += xd_NodeChanged;
			xd.NodeChanged += xd_NodeChanged;
		}

		static void xd_NodeChanged(Object o, XmlNodeChangedEventArgs e)
		{
			var root = ((corpus_xmldoc)o).DocumentElement;
			if (root != null)
				root.Modified = true;
		}

		protected override void OnPropertyChanged(string s)
		{
			base.OnPropertyChanged(s);

			if (s == "FullPath")
			{
				OnPropertyChanged("Directory");
				OnPropertyChanged("Filename");
			}
			if (s == "Filename")
			{
				OnPropertyChanged("CorpusName");
			}
			if (s == "Filename" || s == "Directory")
			{
				OnPropertyChanged("Exists");
			}
			if (s == "Exists")
			{
				OnPropertyChanged("IsReadOnly");
			}
			if (s == "Modified" || s == "Exists")
			{
				OnPropertyChanged("StatusIconKey");
				OnPropertyChanged("NeedSave");
			}
		}

		String _corpus_name;
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public String CorpusName
		{
			get
			{
				return Filename ?? _corpus_name ?? ("Corpus-" + GetHashCode().ToString("X8") + ".xml");
			}
			set
			{
				_corpus_name = value;
			}
		}

		String _fullpath;
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public String FullPath
		{
			get { return _fullpath; }
			set
			{
				if (_fullpath != value)
				{
					_fullpath = value;
					OnPropertyChanged("FullPath");
				}
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public String Directory
		{
			get
			{

				var fn = FullPath;
				return fn != null ? Path.GetDirectoryName(fn) : null;
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public String Filename
		{
			get
			{
				var fn = FullPath;
				return fn != null ? Path.GetFileName(fn) : null;
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool Exists
		{
			get
			{
				var fn = FullPath;
				return fn != null && File.Exists(fn);
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public new bool IsReadOnly
		{
			get
			{
				var fn = FullPath;
				return fn != null && new FileInfo(fn).IsReadOnly;
			}
		}

		bool _modified = true;
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool Modified
		{
			get { return _modified; }
			set
			{
				if (value != _modified)
				{
					_modified = value;
					OnPropertyChanged("Modified");
				}
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public Object StatusIconKey
		{
			get
			{
				if (!Exists)
					return "corpus_status_no_file";
				else if (Modified)
					return "corpus_status_dirty";
				return "corpus_status_clean";
			}
		}

		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public bool NeedSave
		{
			get { return !Exists || Modified; }
		}
	};
}
